import gevent

from .abstract import AbstractBackend
from pipeliner.exceptions import PipelineAlreadyRunning


class GeventBackend(AbstractBackend):
    def __init__(self):
        self.is_running = False
        self._greenlet = None

    def run(self, target, *args, **kwargs):
        if self.is_running:
            raise PipelineAlreadyRunning()

        self.is_running = True
        self._greenlet = gevent.spawn(target, *args, **kwargs)

    def wait_until_complete(self):
        self._greenlet.join()



class RegisterException(Exception):
    pass


class TaskAlreadyRegistered(RegisterException):
    pass


class TaskNotFound(RegisterException):
    pass


class PipelineException(Exception):
    pass


class TaskDependencyError(PipelineException):
    pass


class PipelineAlreadyRunning(PipelineException):
    pass



from inspect import isfunction
from pipeliner.register import Register
from pipeliner.context import Context
from pipeliner.exceptions import TaskDependencyError
from .backend import GeventBackend


class Pipeline(object):
    def __init__(self, first_task, *tasks, **kwargs):
        self._providers = {}
        self._backend = GeventBackend()

        self._check_tasks(first_task, *tasks)

        self._context = Context(**kwargs)

    def _check_tasks(self, first_task, *tasks):
        if len(self._get_depends(first_task)) > 0:
            raise TaskDependencyError("First task can't have dependencies")

        self._tasks = [first_task] + list(tasks)

        for task in self._tasks:
            if not isfunction(task):
                raise ValueError("Task must be a function")
            if not self._is_applied(task):
                raise ValueError(
                    "Task must be applied before adding to pipeline")

            depends = self._get_depends(task)
            for dep in depends:
                if not self._is_dependency_satisfied(dep):
                    raise TaskDependencyError("Dependency `{}` of task `{}` " \
                        "is not satisfied".format(dep, task.__name__))

            provides = self._get_provides(task)
            for prov in provides:
                self._add_provider(prov, task)

    def _is_applied(self, task):
        return hasattr(task, '_applied')

    def _get_depends(self, task):
        return Register().get_depends(task.__name__)

    def _get_provides(self, task):
        return Register().get_provides(task.__name__)

    def _add_provider(self, dependency, task):
        self._providers[dependency] = task

    def _is_dependency_satisfied(self, dependency):
        return (dependency in self._providers)

    def run(self, wait=False):
        self._backend.run(self._run_tasks)
        if wait:
            self.wait_until_complete()

    def _run_tasks(self):
        for task in self._tasks:
            self._context = task(self._context)

    def wait_until_complete(self):
        self._backend.wait_until_complete()

    @property
    def tasks(self):
        return self._tasks



from functools import update_wrapper
from six import with_metaclass
from pipeliner.exceptions import TaskAlreadyRegistered, TaskNotFound


class Singleton(type):
    _instance = None

    def __call__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = type.__call__(cls, *args, **kwargs)
        return cls._instance


class Register(with_metaclass(Singleton, object)):
    """
    Handles task registration (including providers control)
    """

    _handlers = {}

    def add_handler(self, func_name, depends=[], provides=[]):
        if func_name not in self._handlers:
            self._handlers[func_name] = dict(depends=depends, provides=provides)
        else:
            raise TaskAlreadyRegistered()

    def get_handler(self, func_name):
        if func_name in self._handlers:
            return self._handlers[func_name]
        else:
            raise TaskNotFound()

    def _get_handler_data(self, func_name, data):
        return self.get_handler(func_name)[data]

    def get_depends(self, func_name):
        return self._get_handler_data(func_name, 'depends')

    def get_provides(self, func_name):
        return self._get_handler_data(func_name, 'provides')

    def _clear(self):
        """ Remove all handlers """
        self._handlers = {}


def task(depends=[], provides=[]):
    def wrapper(f):
        Register().add_handler(f.__name__, depends=depends, provides=provides)
        def args_wrapper(*args, **kwargs):
            def context_wrapper(context):
                f(context, *args, **kwargs)
                return context
            update_wrapper(context_wrapper, f)
            context_wrapper._applied = True
            return context_wrapper
        update_wrapper(args_wrapper, f)
        return args_wrapper
    return wrapper



from abc import ABCMeta, abstractmethod, abstractproperty
from six import with_metaclass


class AbstractBackend(with_metaclass(ABCMeta, object)):
    @abstractmethod
    def run(self, target, *args, **kwargs):
        pass

    @abstractmethod
    def wait_until_complete(self):
        pass



from setuptools import setup, find_packages


setup(
    name="pipeliner",
    description="Task execution framework with dependencies control",
    version="0.0.3",
    author="Nikolay Sedelnikov",
    author_email="n.sedelnikov@gmail.com",
    packages=find_packages(),
    install_requires=[
        'gevent',
        'six',
    ],
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest',
        'mock',
        'pytest-cov',
    ],
    include_package_data=True,
    zip_safe=True,
)
